(    Title:  Tests for ttester-xf
      File:  ttester-xf-test.fs
  Log file:  ttester-xf.log
    Author:  David N. Williams
   License:  Public Domain
   Version:  1.3.1
   Revised:  February 1, 2011

These tests are not as strict about minimizing the use of non-
CORE words, or non-FLOATING words for floating point, as is
ttester-xf itself.
)
decimal

\ debugging
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]
: .#  ( n -- )  cr ." #" . cr ;

\ Ensure that fp is enabled for pfe in ttester-xf.
s" FLOATING-EXT" environment? [IF] ( flag) drop [ELSE]
cr .( FLOATING EXT NOT AVAILABLE) cr ABORT [THEN]

\ debugging
: .depths  ( -- )
  ."   data depth: " depth .  ."   fp depth: " fdepth . cr ;

 2 CONSTANT MAX-XT-INITIALS
 2 CONSTANT MAX-FT-INITIALS
\ We use values because we want to make these smaller later:
16 VALUE MAX-XT-RESULTS   16 VALUE MAX-FT-RESULTS

#13 BASE !                \ test BASE invariance
 s" ttester-xf.fs" INCLUDED false
\ s" xftester.fs" INCLUDED true
CONSTANT XFTESTER

X{ BASE @ X-> #13 }X
T{ BASE @  -> #13 }T
decimal
true VERBOSE !

s" tester-display.fs" INCLUDED
s" xtester-errors.fs" INCLUDED
s" ftester-errors.fs" INCLUDED

: ?.errors  ( -- )
  blue-text s" Unexpected " ?type ?.xt-errors ?.cr
  blue-text s" Unexpected " ?type ?.ft-errors ;

\ less elaborate errors for now
' XT-ERROR1 XT-ERROR-XT !
' FT-ERROR1 FT-ERROR-XT !

\ before the first displayed line
PFE-HOST IFORTH-HOST or [IF] ?.cr [THEN]

XFTESTER [IF]
  ?." Using xftester.fs
[ELSE]
  ?." Using ttester-xf.fs
[THEN]

PFE-HOST [IF]
  ?." Host is pfe
\ 0 REDEFINED-MSG !
[THEN]
GFORTH-HOST [IF]
  ?." Host is gforth
\ 0 WARNINGS !
[THEN]
IFORTH-HOST [IF]
  ?." Host is iForth
\ 0 WARNING !
[THEN]

\ *** TESTS

\ scratch
VARIABLE v1   FVARIABLE fv1
VARIABLE v2   FVARIABLE fv2
VARIABLE v3

(
The tests in the next two sections are copies of tests in
ttester-xfarray-test.fs that already passed with the gforth cvs
version 1.1.6 of ttester.  Failure would mean a bug in a
component of T{, ->, or }T.
)

TESTING  XT-A<> XT-A@ XT-A!

CREATE []         0 ,
CREATE [0]        1 ,  0 ,
CREATE [10]       1 , 10 ,
CREATE [10,11]    2 , 10 , 11 ,
CREATE [10,1]     2 , 10 ,  1 ,
CREATE [0,11]     2 ,  0 , 11 ,
CREATE [0,1]      2 ,  0 ,  1 ,
CREATE [10,11,12] 3 , 10 , 11 , 12 ,
CREATE [1,11,12]  3 ,  1 , 11 , 12 ,
CREATE [1,1,12]   3 ,  1 ,  1 , 12 ,
CREATE [1,11,2]   3 ,  1 , 11 ,  2 ,

\ different length
T{   []  [10,11,12] XT-A<> -> -1 }T
T{ [10]  [10,1]     XT-A<> -> -1 }T
T{ [10]  [10,11,12] XT-A<> -> -1 }T

\ same length returns index+1 or 0

\ different at index 0
T{ [0]    [10]   XT-A<> -> 1 }T
T{ [10,1] [0,11] XT-A<> -> 1 }T
T{ [10,1] [0,1]  XT-A<> -> 1 }T
T{ [10,11,12] [1,11,12] XT-A<> -> 1 }T
T{ [10,11,12] [1,1,12]  XT-A<> -> 1 }T

\ different at index 1
T{ [10,1] [10,11]     XT-A<> -> 2 }T
T{ [1,1,12] [1,11,12] XT-A<> -> 2 }T
T{ [1,1,12] [1,11,2]  XT-A<> -> 2 }T

\ different at index 2
T{ [1,11,12] [1,11,2] XT-A<> -> 3 }T

\ the same
T{ []         []         XT-A<> -> 0 }T
T{ [10]       [10]       XT-A<> -> 0 }T
T{ [10,11,12] [10,11,12] XT-A<> -> 0 }T

T{ []         XT-A@ -> 0 }T
T{ [10]       XT-A@ -> 10 1 }T
T{ [10,11]    XT-A@ -> 11 10 2 }T
T{ [10,11,12] XT-A@ -> 12 11 10 3 }T

create max3array 4 cells allot
T{ 3  2  1 3 max3array XT-A! max3array XT-A@ -> 3  2  1 3 }T
T{   11 10 2 max3array XT-A! max3array XT-A@ -> 11 10 2 }T
T{       0 1 max3array XT-A! max3array XT-A@ -> 0 1 }T
T{         0 max3array XT-A! max3array XT-A@ -> 0 }T
\ xarray overwrite?
T{ -1 -1 -1 3 max3array XT-A! -> }T
T{     0  0 2 max3array XT-A! -> }T
T{ max3array 2 cells + 2@ -> -1 0 }T

TESTING  SET-FT-MODE-EXACT FT-A<> FT-A@ FT-A!

\ default is exact 
T{ FT-TEST=-XT @ -> ' FT-DATUM= }T

0 FT-TEST=-XT !
T{ SET-FT-MODE-EXACT FT-TEST=-XT @ -> ' FT-DATUM= }T

: FARRAY  ( +n -- ) ( f: xn ... x1 -- )
  CREATE dup , falign here >r floats allot here r>
  ?DO i f! 1 floats +LOOP ;

0             FARRAY f[]
1 0E          FARRAY f[0]
1 10E         FARRAY f[10]
2 11E 10E     FARRAY f[10,11]
2  1E 10E     FARRAY f[10,1]
2 11E  0E     FARRAY f[0,11]
2  1E  0E     FARRAY f[0,1]
3 12E 11E 10E FARRAY f[10,11,12]
3 12E 11E  1E FARRAY f[1,11,12]
3 12E  1E  1E FARRAY f[1,1,12]
3  2E 11E  1E FARRAY f[1,11,2]

T{ f[] @ -> 0 }T
T{ f[10] @ f[10] cell+ faligned f@ -> 1 10E }T
T{ f[10,11] @ f[10,11] cell+ faligned dup f@ float+ f@ -> 2 10e 11e }T

\ different length
T{   f[]  f[10,11,12] FT-A<> -> -1 }T
T{ f[10]  f[10,1]     FT-A<> -> -1 }T
T{ f[10]  f[10,11,12] FT-A<> -> -1 }T

\ same length returns index+1 or 0

\ different at index 0
T{ f[0]    f[10]   FT-A<> -> 1 }T
T{ f[10,1] f[0,11] FT-A<> -> 1 }T
T{ f[10,1] f[0,1]  FT-A<> -> 1 }T
T{ f[10,11,12] f[1,11,12] FT-A<> -> 1 }T
T{ f[10,11,12] f[1,1,12]  FT-A<> -> 1 }T

\ different at index 1
T{ f[10,1] f[10,11]     FT-A<> -> 2 }T
T{ f[1,1,12] f[1,11,12] FT-A<> -> 2 }T
T{ f[1,1,12] f[1,11,2]  FT-A<> -> 2 }T

\ different at index 2
T{ f[1,11,12] f[1,11,2] FT-A<> -> 3 }T

\ the same
T{ f[]         f[]         FT-A<> -> 0 }T
T{ f[10]       f[10]       FT-A<> -> 0 }T
T{ f[10,11,12] f[10,11,12] FT-A<> -> 0 }T

T{ f[]         FT-A@ -> 0 }T
T{ f[10]       FT-A@ -> 10E 1 }T
T{ f[10,11]    FT-A@ -> 11E 10E 2 }T
T{ f[10,11,12] FT-A@ -> 12E 11E 10E 3 }T

here 1 cells allot falign 3 floats allot CONSTANT max3farray
T{ 3E  2E  1E 3 max3farray FT-A! max3farray FT-A@ ->  3E  2E  1E 3 }T
T{    11E 10E 2 max3farray FT-A! max3farray FT-A@ -> 11E 10E 2 }T
T{         0E 1 max3farray FT-A! max3farray FT-A@ ->  0E 1 }T
T{            0 max3farray FT-A! max3farray FT-A@ ->  0 }T
\ farray overwrite?
T{ -1E -1E -1E 3 max3farray FT-A! -> }T
T{      0E  0E 2 max3farray FT-A! -> }T
T{ max3farray cell+ faligned float+ dup f@ float+ f@ -> 0E -1E }T

TESTING  X{ X-> }X

X{ X-> }X
X{ 1 X-> 1 }X
X{ 1 2 3 1 X-> 1 2 3 1 }X

\ initial data stack preservation up to MAX-XT-INITIALS
X{ 3 X-> 3 }X
depth v1 ! X{ v1 @ X-> 0 }X
1 X{ 3 X-> 3 }X
v1 ! depth v2 !
X{ v1 @ v2 @ X-> 1 0 }X
1 2 X{ 3 X-> 3 }X
v1 ! v2 ! depth v3 !
X{ v1 @ v2 @ v3 @ X-> 2 1 0 }X

TESTING  F{ F-> }F

T{ SET-FT-MODE-EXACT -> }T

F{ F-> }F
F{ 1E F-> 1E }F
F{ 1E 2E 3E 1E F-> 1E 2E 3E 1E }F

\ initial fp stack preservation up to MAX-FT-INITIALS
F{ 3E F-> 3E }F
fdepth v1 !
X{ v1 @ X-> 0 }X

1E F{ 3E F-> 3E }F
fv1 f! fdepth v2 !
T{ fv1 f@ v2 @ -> 1E 0 }T

1E 2E F{ 3E F-> 3E }F
fv1 f! fv2 f! fdepth v1 !
T{ fv1 f@ fv2 f@ v1 @ -> 2E 1E 0 }T

TESTING  T{ -> }T

SET-FT-MODE-EXACT

T{ -> }T
T{ 1 -> 1 }T
T{ 1E -> 1E }T
T{ 1 1E -> 1E 1 }T

depth v1 ! fdepth v2 !
T{ v1 @ v2 @ -> 0 0 }T
11 T{ 1 1E -> 1E 1 }T
v1 ! depth v2 ! fdepth v3 !
T{ v1 @ v2 @ v3 @ -> 11 0 0 }T

11E T{ 2 2E -> 2 2E }T
fv1 f! depth v1 ! fdepth v2 !
T{ fv1 f@ v1 @ v2 @ -> 11E 0 0 }T

12E 12 13E 13 T{ 3E 3 -> 3E 3 }T
fv1 f! fv2 f! v1 ! v2 !
T{ fv1 f@ fv2 f@ v1 @ v2 @ -> 13E 12E 13 12 }T
depth v1 ! fdepth v2 !
T{ v1 @ v2 @ -> 0 0 }T

TESTING  FT-TEST= FT-DATUM=

T{ 1E 1.0000000001E FT-DATUM= -> false }T
T{ 1E 1E            FT-DATUM= -> true }T
T{ 1E 1.0000000001E FT-TEST= -> false }T
T{ 1E 1E            FT-TEST= -> true }T

VERBOSE @ [IF]
FT-#ERRORS @
?." Inspect the error line:
T{ SET-FT-MODE-EXACT 1E -> 1.0000000001E }T
FT-#ERRORS !
T{ 1E -> 1E }T
[THEN]

\ Comment this section out for systems without IEEE 754 support. 
true [IF]
T{ 0E fnegate fdup       FT-DATUM= -> true }T
T{ 0E fdup fnegate       FT-DATUM= -> false }T
T{ 1E 0E f/ fdup         FT-DATUM= -> true }T
T{ 1E 0E f/ fdup fnegate FT-DATUM= -> false }T
T{ 0E 0E f/ fdup         FT-DATUM= -> true }T
T{ 0E 0E f/ fdup fnegate FT-DATUM= -> false }T
[THEN]

\ This section assumes systems with IEEE 754 support.
VERBOSE @ [IF]
XT-#ERRORS @   FT-#ERRORS @
?." Inspect 3 error lines:
T{ 0E -> -0E }T
T{ 1E 0E f/ -> -1E 0E f/ }T
T{ 0E 0E f/ -> 0E 0E f/ fnegate }T
FT-#ERRORS !   XT-#ERRORS !
[THEN]

TESTING  FT-ABS= FT-REL= FT-AREL= FT-REL0=
?.(          SET-FT-MODE-ABS SET-FT-MODE-REL SET-FT-MOD-AREL SET-FT-MODE-REL0)
?.cr

: ?.FT-MAX-ERRORS
  VERBOSE @ IF blue-text
    ." Using FT-ABS-ERROR:  " FT-ABS-ERROR  f@ fs. cr
    ."       FT-REL-ERROR:  " FT-REL-ERROR  f@ fs. cr
    ."       FT-AREL-ERROR: " FT-AREL-ERROR f@ fs. cr
    normal-text THEN ;

.0005E                    FCONSTANT TEST-ABS-ERR
1.001E 1E f- 1E f/        FCONSTANT TEST-REL-ERR
1.0001E 1E f- 1.00005E f/ FCONSTANT TEST-AREL-ERR

TEST-ABS-ERR  FT-ABS-ERROR f!
TEST-REL-ERR  FT-REL-ERROR f!
TEST-AREL-ERR FT-AREL-ERROR f!
?.FT-MAX-ERRORS

T{ 1.00025E 1E FT-ABS= -> true }T
T{ 1.001E 1E   FT-ABS= -> false }T
T{ 1.0009E10 1E10 FT-REL= -> true }T
T{ 1.0011E10 1E10 FT-REL= -> false }T
\ next two would be opposite if meas and ref were backwards
IFORTH-HOST 0= [IF]
T{ 1.001E10  1E10 FT-REL= -> false }T  \ must be less, not equal to eps
[ELSE]
\ decimal conversion roundoff workaround
T{ 1.00100000000000000001E10 1E10 FT-REL= -> false }T
[THEN]
T{ 1E10 1.001E10 FT-REL= -> true }T
T{ 1.00009E12 1E12 FT-AREL= -> true }T
T{ 1.00011E12 1E12 FT-AREL= -> false }T

T{ SET-FT-MODE-ABS 1.00025E -> 1E }T
T{ FT-TEST=-XT @ -> ' FT-ABS= }T
T{ SET-FT-MODE-REL 1.0009E10 -> 1E10 }T
T{ FT-TEST=-XT @ -> ' FT-REL= }T
T{ SET-FT-MODE-AREL 1.00009E12 -> 1E12 }T
T{ FT-TEST=-XT @ -> ' FT-AREL= }T
T{ SET-FT-MODE-REL0 1.00025E 1.0009E10 -> 1E 1E10 }T
T{ FT-TEST=-XT @ -> ' FT-REL0= }T

VERBOSE @ [IF]
FT-#ERRORS @
?." Inspect 5 error lines:
T{ SET-FT-MODE-ABS  1.001E     -> 1E }T
\ next tolerance fails because it's equal, not less
T{ SET-FT-MODE-REL  1.001E10   -> 1E10 }T
T{ SET-FT-MODE-AREL 1.00011E12 -> 1E12 }T
T{ SET-FT-MODE-REL0 1.001E     -> 1E }T
T{ SET-FT-MODE-REL0 1.001E10   -> 1E10 }T
FT-#ERRORS !
[THEN]

TESTING  FT~ABS= FT~REL= FT-NEARLY= FT-CONF= SET-EXACT SET-NEAR
(
These are tests for ttester compatibility.
)

T{ SET-EXACT 1E 1.0000000001E FT-CONF= -> false }T
T{ FT-TEST=-XT @ FT-EXACT? -> ' FT-CONF= true }T
T{ 1E 1E FT-CONF= -> true }T
T{ 1E -> 1E }T

TEST-AREL-ERR 2E f/ FCONSTANT TEST-REL-NEAR

TEST-ABS-ERR  ABS-NEAR f!
TEST-REL-NEAR REL-NEAR f!

T{ SET-NEAR 1.00025E -> 1E }T
T{ FT-TEST=-XT @ FT-EXACT? -> ' FT-CONF= false }T
T{ 1.00025E 1E FT~ABS= -> true }T
T{ 1.001E 1E   FT~ABS= -> false }T
T{ 1.00009E12 1E12 FT~REL= -> true }T
T{ 1.00011E12 1E12 FT~REL= -> false }T

SET-NEAR
0E ABS-NEAR f!   0E REL-NEAR f!
T{ 1E -> 1E }T
TEST-ABS-ERR ABS-NEAR f!   0E REL-NEAR f!
T{ 1E -> 1E }T
T{ 1.00025E -> 1E }T
0E ABS-NEAR f!   TEST-REL-NEAR REL-NEAR f!
T{ 1E -> 1E }T
T{ 1.00009E12 -> 1E12 }T

VERBOSE @ [IF]
FT-#ERRORS @
?." Inspect 4 error lines:
T{ SET-EXACT 1E -> 1.0000000001E }T
0E ABS-NEAR f!   0E REL-NEAR f!
T{ SET-NEAR  1E -> 1.0000000001E }T \ exact
TEST-ABS-ERR ABS-NEAR f!   0E REL-NEAR f!
T{ SET-NEAR  1.001E -> 1E }T        \ abs
0E ABS-NEAR f!   TEST-REL-NEAR REL-NEAR f!
T{ SET-NEAR  1.00011E12 -> 1E12 }T  \ rel
FT-#ERRORS !
[THEN]

TESTING  exceptions

[UNDEFINED] drops [IF]
  : drops  ( xn ... x1 n -- )  0 ?DO drop LOOP ;
[THEN]

[UNDEFINED] fdrops [IF]
  : fdrops  ( f: xn ... x1 -- ) ( n -- )  0 ?DO fdrop LOOP ;
[THEN]

\ MAX-XT-INITIALS overflow

\ 1 2 3 T{ 
1 2 3 ' T{ CATCH v1 ! 3 drops depth v2 !
T{ v1 @ v2 @ -> -2 0 }T

\ MAX-XT-RESULTS overflow

2 to MAX-XT-RESULTS
\ 1 2 3 ->
1 2 3 ' -> CATCH v1 ! 3 drops depth v2 !
T{ v1 @ v2 @ -> -2 0 }T

\ 1 2 3 }T
1 2 3 ' }T CATCH v1 ! 3 drops depth v2 !
T{ v1 @ v2 @ -> -2 0 }T

\ MAX-FT-INITIALS overflow

\ 1E 2E 3E T{
X{ 1E 2E 3E ' F{ CATCH 3 fdrops fdepth X-> -2 0 }X

\ MAX-FT-RESULTS overflow

1 to MAX-FT-RESULTS
\ 1E 2E ->
X{ 1E 2E ' F-> CATCH 2 fdrops fdepth X-> -2 0 }X

\ 1E 2E }T
X{ 1E 2E ' }F CATCH 2 fdrops fdepth X-> -2 0 }X

\ There's no portable way to CATCH these underflows.  Uncomment
\ one at a time.
\ X{ drop
\ X-> drop
\ F{ fdrop
\ F-> fdrop

TESTING  errors

2 to MAX-FT-RESULTS
VERBOSE @ [IF]

XT-#ERRORS @  FT-#ERRORS @
?." Inspect the error lines:
T{ 1 ->   }T
T{   -> 1 }T
T{ 1E ->    }T
T{    -> 1E }T
T{ 1 -> 2 }T
T{ 1E -> 2E  }T
T{ 1 -> 2 3E }T
T{ 2 3E -> 1 }T
T{ 1E -> 2E 3 }T
T{ 2E 3 -> 1E }T

?." Inspect the more elaborate error reports
' XT-ERROR3 XT-ERROR-XT !
' FT-ERROR3 FT-ERROR-XT !
T{ 1E 2E -> 1E 3E 3 4 }T
T{ 1 2 -> 3 2 }T
3 to MAX-FT-RESULTS
T{ 1E 2E 3E -> 4E 2E 3E }T
T{ 1E -> 2E }T
' XT-ERROR1 XT-ERROR-XT !
' FT-ERROR1 FT-ERROR-XT !

?." The following are actually syntax errors:
T{ 1 }T
T{ 1E }T
FT-#ERRORS ! XT-#ERRORS !
[THEN]  \ VERBOSE

?." This syntax error is not reported: T{ }T
T{ }T

?.errors GFORTH-HOST [IF] ?.cr [THEN]
