(    Title:  Tests for ttester-xf arrays
      File:  ttester-xfarray-test.fs
  Log File:  ttester-xf.log
    Author:  David N. Williams
   License:  Public Domain
   Version:  1.3.1
   Revised:  February 5, 2011

This file tests the array factors in ttester-xf.  It should be
tested with an "original" version of ttester, one without
explicit array words, as a partial bootstrap for testing
ttester-xf with itself.
)
decimal

\ debugging
[UNDEFINED] \\ [IF]
: \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]

\ Ensure that fp is enabled for pfe in ttester.
s" FLOATING-EXT" environment? [IF] ( flag) drop [ELSE]
cr .( FLOATING EXT NOT AVAILABLE) cr ABORT [THEN]

\ debugging
: .depths  ( -- )
  ."   data depth: " depth .  ."   fp depth: " fdepth . cr ;
: .#  ( n -- )  cr ." #" . cr ;

s" ttester.fs" INCLUDED
s" tester-display.fs" INCLUDED
true VERBOSE !
\ before the first displayed line
PFE-HOST IFORTH-HOST or [IF] ?.cr [THEN]

PFE-HOST [IF]
  ?." Host is pfe
\ 0 REDEFINED-MSG !
[THEN]
GFORTH-HOST [IF]
  ?." Host is gforth
\ 0 WARNINGS !
[THEN]
IFORTH-HOST [IF]
  ?." Host is iForth
\ 0 WARNING !
[THEN]


\ *** TTESTER ERROR DISPLAY

VARIABLE #errors    0 #errors !
:noname  ( c-addr u -- )
\ Display an error message followed by the line that had the
\ error.
  red-text 1 #errors +! ERROR1 normal-text ; ERROR-XT !

: ?.errors  ( -- )
  VERBOSE @
  IF blue-text ." #ERRORS: " normal-text #errors @ . THEN ;


\ *** XT-ARRAY WORDS
(
The first array slot has index zero.
)

\ overflow checking to be done elsewhere
: XT-A!  ( x_[n-1] ... x_0 +n addr -- )
  OVER 1+ CELLS OVER + SWAP DO I ! 1 CELLS +LOOP ;

: XT-A@  ( addr -- x_[n-1] ... x_0 +n )
  DUP @ CELLS OVER + DO I @ -1 CELLS +LOOP ;

[UNDEFINED] NIP [IF] : NIP  ( a b -- b)  SWAP DROP ; [THEN]

: XT-A<>  ( addr1 addr2 -- [index+1]|-1|0 )
(
Nonzero return for errors:
  -1       unequal array size
  index+1  deepest unequal array elements
)
  DUP @ 2>R    ( addr1 r: addr2 #left )
  DUP @ R@ <> IF 2R> 2DROP DROP -1 EXIT THEN
  ( addr1) DUP
  BEGIN  CELL+ R@    ( addr1 addr1' #left)
  WHILE  R> R> CELL+ ( addr1 addr1' #left addr2')
         DUP >R SWAP 1- >R @ OVER @ =
  WHILE
  REPEAT ( addr1 addr1') DROP @ ( len) 2R> NIP - ( index+1)
  ELSE   ( addr1 addr1') 2DROP 2R> 2DROP 0
  THEN ;


\ *** XT-ARRAY TESTS

TESTING  XT-A<> XT-A@ XT-A!

CREATE []         0 ,
CREATE [0]        1 ,  0 ,
CREATE [10]       1 , 10 ,
CREATE [10,11]    2 , 10 , 11 ,
CREATE [10,1]     2 , 10 ,  1 ,
CREATE [0,11]     2 ,  0 , 11 ,
CREATE [0,1]      2 ,  0 ,  1 ,
CREATE [10,11,12] 3 , 10 , 11 , 12 ,
CREATE [1,11,12]  3 ,  1 , 11 , 12 ,
CREATE [1,1,12]   3 ,  1 ,  1 , 12 ,
CREATE [1,11,2]   3 ,  1 , 11 ,  2 ,

\ different length
T{   []  [10,11,12] XT-A<> -> -1 }T
T{ [10]  [10,1]     XT-A<> -> -1 }T
T{ [10]  [10,11,12] XT-A<> -> -1 }T

\ same length returns index+1 or 0

\ different at index 0
T{ [0]    [10]   XT-A<> -> 1 }T
T{ [10,1] [0,11] XT-A<> -> 1 }T
T{ [10,1] [0,1]  XT-A<> -> 1 }T
T{ [10,11,12] [1,11,12] XT-A<> -> 1 }T
T{ [10,11,12] [1,1,12]  XT-A<> -> 1 }T

\ different at index 1
T{ [10,1] [10,11]     XT-A<> -> 2 }T
T{ [1,1,12] [1,11,12] XT-A<> -> 2 }T
T{ [1,1,12] [1,11,2]  XT-A<> -> 2 }T

\ different at index 2
T{ [1,11,12] [1,11,2] XT-A<> -> 3 }T

\ the same
T{ []         []         XT-A<> -> 0 }T
T{ [10]       [10]       XT-A<> -> 0 }T
T{ [10,11,12] [10,11,12] XT-A<> -> 0 }T

T{ []         XT-A@ -> 0 }T
T{ [10]       XT-A@ -> 10 1 }T
T{ [10,11]    XT-A@ -> 11 10 2 }T
T{ [10,11,12] XT-A@ -> 12 11 10 3 }T

create max3array 4 cells allot
T{ 3  2  1 3 max3array XT-A! max3array XT-A@ -> 3  2  1 3 }T
T{   11 10 2 max3array XT-A! max3array XT-A@ -> 11 10 2 }T
T{       0 1 max3array XT-A! max3array XT-A@ -> 0 1 }T
T{         0 max3array XT-A! max3array XT-A@ -> 0 }T
\ xarray overwrite?
T{ -1 -1 -1 3 max3array XT-A! -> }T
T{     0  0 2 max3array XT-A! -> }T
T{ max3array 2 cells + 2@ -> -1 0 }T


DEPTH 1E DEPTH 1- FDROP = 0=
[IF] .( FLOATING-STACK REQUIRED) ABORT [THEN]

\ *** FT-ARRAY WORDS
(
The first array slot has index zero.
)

: FT-TEST=  FCONF= ;

\ Overflow checking to be done elsewhere.
: FT-A!   ( +n addr -- ) ( f: r_[n-1] ... r_0 -- )
  2DUP ! CELL+ FALIGNED SWAP ( f-addr n) DUP
  IF   FLOATS OVER + SWAP
       DO I F! 1 FLOATS +LOOP 
  ELSE 2DROP THEN ;

: FT-A@   ( addr -- +n ) ( f: -- r_[n-1] ... r_0 )
  DUP @ DUP >R ( addr n r: n)
  IF   CELL+ FALIGNED R@ 1- FLOATS OVER + 
       DO I F@ -1 FLOATS +LOOP
  ELSE DROP THEN R> ;

[UNDEFINED] NIP [IF] : NIP  ( a b -- b)  SWAP DROP ; [THEN]

: FT-A<>  ( addr1 addr2 -- [index+1]|-1|0 )
(
Nonzero return for errors:
  -1       unequal array size
  index+1  deepest unequal array elements
)
  DUP @ SWAP CELL+ FALIGNED >R >R  ( addr1 r: f-addr2 #left )
  DUP @ R@ <> IF 2R> 2DROP DROP -1 EXIT THEN
  ( addr1) DUP CELL+ FALIGNED
  BEGIN  R@ ( addr1 f-addr1 #left)
  WHILE  ( f-addr1) DUP F@ FLOAT+
         R> ( #left) 1- R>
         ( f-addr2) DUP F@ FLOAT+ >R ( #left) >R
         FT-TEST=
  WHILE
  REPEAT ( addr1 f-addr1) DROP @ ( len) 2R> NIP - ( index+1)
  ELSE   ( addr1 f-addr1) 2DROP 2R> 2DROP 0
  THEN ;


\ *** FT-ARRAY TESTS

TESTING  FT-A<> FT-A@ FT-A!

SET-EXACT

: FARRAY  ( +n -- ) ( f: x_[n+1] ... x_0 -- )
  CREATE dup , falign here >r floats allot here r>
  ?DO i f! 1 floats +LOOP ;

0             FARRAY f[]
1 0E          FARRAY f[0]
1 10E         FARRAY f[10]
2 11E 10E     FARRAY f[10,11]
2  1E 10E     FARRAY f[10,1]
2 11E  0E     FARRAY f[0,11]
2  1E  0E     FARRAY f[0,1]
3 12E 11E 10E FARRAY f[10,11,12]
3 12E 11E  1E FARRAY f[1,11,12]
3 12E  1E  1E FARRAY f[1,1,12]
3  2E 11E  1E FARRAY f[1,11,2]

T{ f[] @ -> 0 }T
T{ f[10] @ f[10] cell+ faligned f@ -> 1 10E }T
T{ f[10,11] @ f[10,11] cell+ faligned dup f@ float+ f@ -> 2 10e 11e }T

\ different length
T{   f[]  f[10,11,12] FT-A<> -> -1 }T
T{ f[10]  f[10,1]     FT-A<> -> -1 }T
T{ f[10]  f[10,11,12] FT-A<> -> -1 }T

\ same length returns index+1 or 0

\ different at index 0
T{ f[0]    f[10]   FT-A<> -> 1 }T
T{ f[10,1] f[0,11] FT-A<> -> 1 }T
T{ f[10,1] f[0,1]  FT-A<> -> 1 }T
T{ f[10,11,12] f[1,11,12] FT-A<> -> 1 }T
T{ f[10,11,12] f[1,1,12]  FT-A<> -> 1 }T

\ different at index 1
T{ f[10,1] f[10,11]     FT-A<> -> 2 }T
T{ f[1,1,12] f[1,11,12] FT-A<> -> 2 }T
T{ f[1,1,12] f[1,11,2]  FT-A<> -> 2 }T

\ different at index 2
T{ f[1,11,12] f[1,11,2] FT-A<> -> 3 }T

\ the same
T{ f[]         f[]         FT-A<> -> 0 }T
T{ f[10]       f[10]       FT-A<> -> 0 }T
T{ f[10,11,12] f[10,11,12] FT-A<> -> 0 }T

T{ f[]         FT-A@ -> 0 }T
T{ f[10]       FT-A@ -> 10E 1 }T
T{ f[10,11]    FT-A@ -> 11E 10E 2 }T
T{ f[10,11,12] FT-A@ -> 12E 11E 10E 3 }T

here 1 cells allot falign 3 floats allot CONSTANT max3farray
T{ 3E  2E  1E 3 max3farray FT-A! max3farray FT-A@ ->  3E  2E  1E 3 }T
T{    11E 10E 2 max3farray FT-A! max3farray FT-A@ -> 11E 10E 2 }T
T{         0E 1 max3farray FT-A! max3farray FT-A@ ->  0E 1 }T
T{            0 max3farray FT-A! max3farray FT-A@ ->  0 }T
\ farray overwrite?
T{ -1E -1E -1E 3 max3farray FT-A! -> }T
T{      0E  0E 2 max3farray FT-A! -> }T
T{ max3farray cell+ faligned float+ dup f@ float+ f@ -> 0E -1E }T

?.errors GFORTH-HOST [IF] ?.cr [THEN]
