(      Title:  A utility for testing Forth
               data and floating-point stack effects
        File:  xftester.fs
  Test files:  ttester-xf-test.fs, ttester-xfarray-test.fs 
    Log file:  ttester-xf.log
     License:  John Hayes, Public Domain
     Version:  1.3.1
     Revised:  January 31, 2011

For any code derived from John Hayes' tester program:
)
\  (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\  MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
The rest is public domain.

This file revises the code for ttester, a utility for testing
Forth words, as developed by several authors [see below].  The
following is a quote from ttester.
)
\ ttester is based on the original tester suite by Hayes:
\ From: John Hayes S1I
\ Subject: tester.fr
\ Date: Mon, 27 Nov 95 13:10:09 PST  
\ (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\ MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
\ VERSION 1.1
\ All the subsequent changes have been placed in the public
\ domain.  The primary changes from the original are the
\ replacement of "{" by "T{" and "}" by "}T" (to avoid conflicts
\ with the uses of { for locals and } for FSL arrays),
\ modifications so that the stack is allowed to be non-empty
\ before T{ , and extensions for the handling of floating point
\ tests.  Code for testing equality of floating point values
\ comes from ftester.fs written by David N. Williams, based on
\ the idea of approximate equality in Dirk Zoller's float.4th.
\ Further revisions were provided by Anton Ertl, including the
\ ability to handle either integrated or separate floating point
\ stacks.  Revision history and possibly newer versions can be
\ found at

\ http://www.complang.tuwien.ac.at/cvsweb/cgi-bin/cvsweb/gforth/test/ttester.fs

\ Explanatory material and minor reformatting (no code changes)
\ by C. G. Montgomery March 2009, with helpful comments from
\ David Williams and Krishna Myneni.
(
In the following, the file ftester.fs is different from that in
the quote above, which has been renamed as ftester-old.fs.

This file is a driver, which loads xtester.fs and ftester.fs. It
is functionally equivalent to ttester-xf.fs, which basically
copies the text of the two files instead of loading them. 
Generic properties of ttester-xf described in the following
apply equally well to xftester.

The designation "x" stands for the generic ANS/ISO Forth
"unspecified cell" data type.

The following revision of the ttester explanatory material, by
David N. Williams, takes into account ttester-xf restrictions
and extras.  Versions 1.2.0 and later and 1.3.0 and later add
new fp comparisons, based on suggestions by Krishna Myneni in
comp.lang.forth.

The basic restriction is that an integrated floating-point
stack is no longer allowed.

A number of words in ttester could be classified as implicitly
private, although a specific user interface was never declared.
In this file the designations PUBLIC and PRIVATE indicate word
sets expected to be supported or not necessarily supported in
future versions.  This is purely a textual affectation, without
actual wordlist management such as that provided by the FSL
PUBLIC: and PRIVATE: words.

Users of ttester who restrict themselves to a separate floating-
point stack, and who use only the following words, can use
ttester-xf as a drop-in replacement:

  T{  ->  }T  VERBOSE  TESTING
  HAS-FLOATING  ABS-NEAR  REL-NEAR  SET-EXACT  SET-NEAR

The ttester words ERROR-XT and ERROR1 are no longer available,
but have been replaced by stack-specific words, XT-ERROR-XT and
FT-ERROR-XT, respectively, XT-ERROR-DEFAULT and
FT-ERROR-DEFAULT.

The following files contain enhancements over default error
reporting:  tester-display.fs, xtester-errors.fs, and
ftester-errors.fs.  See ttester-xf-test.fs for examples of their
use.

These words are currently declared to be the ttester-xf user
interface:

PUBLIC WORDS

   tests:  T{  ->  }T  VERBOSE  TESTING  HAS-FLOATING
           X{  X->  }X  F{  F->  }F
           FT-ABS-ERROR  FT-REL-ERROR  FT-AREL-ERROR
           FT-TEST=-XT
           FT-DATUM=  FT-ABS=  FT-REL=  FT-AREL=  FT-REL0=
           SET-FT-MODE-EXACT  SET-FT-MODE-ABS
           SET-FT-MODE-REL  SET-FT-MODE-AREL  SET-FT-MODE-REL0

           for ttester compatibiity
           ABS-NEAR   REL-NEAR
           SET-EXACT  SET-NEAR

  errors:  XT-ERROR-XT  XT-ERROR-DEFAULT  XT-ERROR-INDEX
           FT-ERROR-XT  FT-ERROR-DEFAULT  FT-ERROR-INDEX

  arrays:  XT-INITIALS  XT-RESULTS  XT-GOALS
           FT-INITIALS  FT-RESULTS  FT-GOALS
           XT-A@  XT-A!  XT-A<>
           FT-A@  FT-A!  FT-A<>

  config:  MAX-XT-INITIALS  MAX-XT-RESULTS
           MAX-FT-INITIALS  MAX-FT-RESULTS

Basic usage takes the form:

  <xfinitials> T{ <xfresults> -> <xfgoals> }T <xfinitials>

Here <xfinitials> are possibly nonempty data and fp stack
contents at the start of the test.  The sequence saves
<xfinitials> and removes them from the stacks, executes the code
that produces the data and fp stack <xfresults>, saves and
removes the <xfresults> from the stacks, executes the code that
produces the data and fp stack <xfgoals>, saves and removes
them, compares the saved <xfresults> and <xfgoals>, reports when
there is a discrepancy between the two, and restores the saved
<xfinitials>.

For example:

T{ 1 2 3 swap -> 1 3 2 }T  ok
T{ 1 2 3 swap -> 1 2 2 }T INCORRECT RESULT: T{ 1 2 3 swap -> 1 2 2 }T ok
T{ 1 2 3 swap -> 1 2 }T WRONG NUMBER OF RESULTS: T{ 1 2 3 swap -> 1 2 }T ok

More examples can be found in the file ttester-xf-test.fs.

In addition to the T{..->..}T and F{..F->..}F tests in ttester
[where "}F" was formerly named "F}"], ttester-xf makes
X{..X->..}X tests available.  The X and F tests act only on the
data and fp stacks, respectively.

Floating-point testing can involve further complications.  The
code attempts to determine whether floating-point support is
present, and behaves accordingly.  The constant HAS-FLOATING
contains the results of its efforts, so the behavior of the test
code can be modified by the user if necessary.

Then there are the perennial issues of floating-point value
comparisons.  Exact equality is specified by SET-FT-MODE-EXACT
[the default].  If one of the included approximate equality
tests is desired, execute SET-FT-MODE-ABS, SET-FT-MODE-REL,
SET-FT-MODE-AREL, or SET-FT-MODE-REL0.  Corresponding to these,
the public fvariables FT-ABS [default 0E], FT-REL [default
1E-12], and FT-AREL [default 1E-12] contain the values to be
used in comparisons by the private words FT-ABS=, FT-REL=, and
FT-AREL=.  FT-REL0= conditionally uses both FT-ABS and FT-REL. 
Here is an example:

  SET-FT-MODE-REL
  1E-6 FT-REL F!
  F{ S" 3.14159" >FLOAT F-> -1E FACOS TRUE }F

Note that comparisons in this mode are not symmetric.  Measured
values must be in the fresults, and reference values in the
fgoals.

The user may supply a different equality test by storing
his own xt in the public variable FT-TEST=-XT.

Storage of initial, result, and goal data includes overflow
checking.  The user can change the overflow limits by defining
one or both of the following constants *before* loading this
file:

  MAX-XT-INITIALS  MAX-XT-RESULTS	
  MAX-FT-INITIALS  MAX-FT-RESULTS

The xt's and tolerances in the following variables and
fvariables can only be changed after loading this file:

  XT-ERROR-XT  FT-ERROR-XT  FT-TEST=-XT
  FT-ABS  FT-REL  FT-AREL
  ABS-NEAR  REL-NEAR  [ttester compatibility]

As with ttester, loading ttester-xf does not change BASE.
Remember that floating-point input is ambiguous if the base is
not decimal.

This version of ttester-xf uses a few more non-CORE words, and
non-FLOATING words in the floating-point sections, than ttester,
namely,  2>R, 2R>, [UNDEFINED], and INCLUDED.  Here is the
complete list:

  TOOLS EXT:      [IF]  [ELSE]  [THEN]  [UNDEFINED]
  CORE EXT:       2>R  2R>
                  TO  VALUE  [ftester only]
  FILE:           (  INCLUDED
  FLOATING EXT:   F~
)

S" xtester.fs" INCLUDED

: "FLOATING" S" FLOATING" ;    \ only compiled S" in CORE
"FLOATING" ENVIRONMENT? [IF]
   [IF] TRUE [ELSE] FALSE [THEN]
[ELSE]
   FALSE
[THEN] CONSTANT HAS-FLOATING

HAS-FLOATING [IF]
S" ftester.fs" INCLUDED
[THEN]

\ Set the following flag to TRUE for more verbose output; this may
\ allow you to tell which test caused your system to hang.

  VARIABLE VERBOSE   FALSE VERBOSE !

: TESTING	\ ( -- ) TALKING COMMENT.
  SOURCE VERBOSE @
  IF DUP >R TYPE CR R> >IN !
  ELSE >IN ! DROP
  THEN ;

\ In the following specifications, references to the fp stack
\ apply when HAS-FLOATING is true.

: T{  ( <xfinitials> -- )
(
Save the data and fp stack <xfinitials> for restoration by }T.
Clear <xfresults> and <xfgoals> storage to put it in a defined
state for }T in case -> is missing.
)
  X{ [ HAS-FLOATING ] [IF] F{ [THEN] ;

: ->  ( <xfresults> -- )
(
Record the depth and contents of the data and fp stacks.
)
  X-> [ HAS-FLOATING ] [IF] F-> [THEN] ;

: }T  ( <xfgoals> -- <xfinitials> )
(
Save the data and fp stack <xfgoals>, compare them with the
saved <xfresults>, and restore the <xfinitials>.
)
  }X [ HAS-FLOATING ] [IF] }F [THEN] ;

\ END of xftester.fs
